@@ -54,6 +54,8 @@ command_input(struct confmodule *mod, char *arg)
 
     if (visible)
         visible = mod->frontend->methods.add(mod->frontend, q);
+    else
+        mod->frontend->methods.add_noninteractive(mod->frontend, q);
 
     if (q->priority != NULL)
         free(q->priority);
@@ -127,9 +129,12 @@ command_capb(struct confmodule *mod, char *arg)
             mod->frontend->capability |= DCF_CAPB_BACKUP;
         else if (strcmp(argv[i], "progresscancel") == 0)
             mod->frontend->capability |= DCF_CAPB_PROGRESSCANCEL;
+        else if (strcmp(argv[i], "align") == 0)
+            mod->frontend->capability |= DCF_CAPB_ALIGN;
     }
 
-    if (asprintf(&out, "%u multiselect backup progresscancel", CMDSTATUS_SUCCESS) == -1)
+    if (asprintf(&out, "%u multiselect backup progresscancel align",
+                 CMDSTATUS_SUCCESS) == -1)
         DIE("Out of memory");
 
     plugin_state = NULL;
@@ -148,7 +153,7 @@ command_capb(struct confmodule *mod, char *arg)
         out = newout;
         outend = mempcpy(outend, " plugin-", 8);
         outend = mempcpy(outend, plugin->name, namelen);
-        *outend++ = '\0';
+        *outend = '\0';
     }
 
     return out;
@@ -159,9 +164,6 @@ command_title(struct confmodule *mod, char *arg)
 {
     char *out;
 
-    if (*arg)
-        debug_printf(0, "Obsolete command TITLE %s called", arg);
-
     mod->frontend->methods.set_title(mod->frontend, arg);
     asprintf(&out, "%u", CMDSTATUS_SUCCESS);
     return out;
@@ -193,30 +195,46 @@ command_go(struct confmodule *mod, char *arg)
     char *out;
     const char *running_frontend = NULL;
     const char *requested_frontend = NULL;
-    struct question *q;
     int ret;
 
     argc = strcmdsplit(arg, argv, DIM(argv) - 1);
     CHECKARGC(== 0);
 
-    q = mod->questions->methods.get(mod->questions, "debconf/frontend");
-    if (q)
-	requested_frontend = question_getvalue(q, "C");
-    question_deref(q);
+    requested_frontend = getenv("DEBIAN_FRONTEND");
 
-    running_frontend = getenv("DEBIAN_FRONTEND");
+    if (requested_frontend)
+    {
+        if (mod != NULL && mod->frontend != NULL)
+            running_frontend = mod->frontend->name;
 
-    if (requested_frontend && strcmp(running_frontend, requested_frontend) != 0) {
-	q = mod->frontend->questions;
-	mod->frontend->methods.shutdown(mod->frontend);
-	if (mod->frontend->handle != NULL)
-	    dlclose(mod->frontend->handle);
-	DELETE(mod->frontend);
-	setenv("DEBIAN_FRONTEND",requested_frontend,1);
-	mod->frontend = frontend_new(mod->config, mod->templates, mod->questions);
-	mod->frontend->questions = q;
+        if (requested_frontend && running_frontend &&
+            strcmp(running_frontend, requested_frontend) != 0)
+        {
+            struct frontend *old_frontend;
+            struct frontend *new_frontend;
+
+            /* Shut down the current frontend, but do not delete unless opening
+             the new one actually succeeds */
+            mod->frontend->methods.shutdown(mod->frontend);
+            
+            new_frontend = frontend_new(mod->config, mod->templates, mod->questions);
+            if (new_frontend)
+            {
+                old_frontend = mod->frontend;
+                mod->frontend = new_frontend;
+                new_frontend->questions = old_frontend->questions;
+                new_frontend->questions_noninteractive = old_frontend->questions_noninteractive;
+                frontend_delete(old_frontend);
+            }
+            else
+            {
+                /* If the new frontend failed, reinitialise the old one */
+                mod->frontend->methods.initialize(mod->frontend, mod->config);
+            }
+        }
     }
 
+    mod->frontend->methods.go_noninteractive(mod->frontend);
     ret = mod->frontend->methods.go(mod->frontend);
     if (ret == CMDSTATUS_GOBACK || mod->backed_up != 0)
     {
@@ -287,27 +305,8 @@ command_set(struct confmodule *mod, char *arg)
         char *prev = STRDUP(question_getvalue(q, ""));
         question_setvalue(q, argv[1]);
 
-        if (mod->questions->methods.set(mod->questions, q) != 0)
-        {
+        if (frontend_qdb_set(mod->questions, q, prev))
             asprintf(&out, "%u value set", CMDSTATUS_SUCCESS);
-            if (0 == strcmp("debconf/language", argv[0]))
-            { /* Pass the value on to getlanguage() in templates.c */
-                debug_printf(0, "Setting debconf/language to %s", argv[1]);
-                setenv("LANGUAGE", argv[1], 1);
-                /* Reloading takes a little while, so only do it when it's
-                 * really necessary.
-                 */
-                if (!load_all_translations() &&
-                    strcmp(argv[1], "C") != 0 && strcmp(argv[1], "en") != 0 &&
-                    (prev == NULL || strcmp(argv[1], prev) != 0))
-                    mod->templates->methods.reload(mod->templates);
-            }
-            else if (strcmp(argv[0], "debconf/priority") == 0)
-            {
-                debug_printf(0, "Setting debconf/priority to %s", argv[1]);
-                setenv("DEBCONF_PRIORITY", argv[1], 1);
-            }
-        }
         else
             asprintf(&out, "%u cannot set value", CMDSTATUS_INTERNALERROR);
 
@@ -463,7 +462,7 @@ command_metaget(struct confmodule *mod, char *arg)
         return out;
     }
 
-    value = question_get_field(q, "", argv[1]);
+    value = question_get_raw_field(q, "", argv[1]);
     if (value == NULL)
         asprintf(&out, "%u %s does not exist", CMDSTATUS_BADQUESTION, argv[1]);
     else
@@ -622,7 +621,6 @@ command_progress(struct confmodule *mod, char *arg)
 {
     int min, max;
     struct question *q = NULL;
-    char *value;
     char *argv[6];
     int argc;
     char *out;
@@ -652,17 +650,9 @@ command_progress(struct confmodule *mod, char *arg)
                     CMDSTATUS_BADQUESTION, argv[3]);
             return out;
         }
-        value = question_get_field(q, "", "description");
-        question_deref(q);
-        if (value == NULL)
-        {
-            asprintf(&out, "%u %s description field does not exist",
-                    CMDSTATUS_BADQUESTION, argv[3]);
-            return out;
-        }
         mod->frontend->methods.progress_start(mod->frontend,
-                min, max, value);
-        free(value);
+                min, max, q);
+        question_deref(q);
     }
     else if (strcasecmp(argv[0], "set") == 0)
     {
@@ -696,16 +686,8 @@ command_progress(struct confmodule *mod, char *arg)
                     CMDSTATUS_BADQUESTION, argv[1]);
             return out;
         }
-        value = question_get_field(q, "", "description");
+        ret = mod->frontend->methods.progress_info(mod->frontend, q);
         question_deref(q);
-        if (value == NULL)
-        {
-            asprintf(&out, "%u %s description field does not exist",
-                    CMDSTATUS_BADQUESTION, argv[1]);
-            return out;
-        }
-        ret = mod->frontend->methods.progress_info(mod->frontend, value);
-        free(value);
         if (ret == DC_GOBACK)
         {
             asprintf(&out, "%u progress bar cancelled",
@@ -717,6 +699,12 @@ command_progress(struct confmodule *mod, char *arg)
     {
         mod->frontend->methods.progress_stop(mod->frontend);
     }
+    else
+    {
+        asprintf(&out, "%u unknown subcommand %s",
+                CMDSTATUS_SYNTAXERROR, argv[0]);
+        return out;
+    }
 
     asprintf(&out, "%u OK", CMDSTATUS_SUCCESS);
     return out;
@@ -739,7 +727,7 @@ command_settitle(struct confmodule *mod, char *arg)
         asprintf(&out, "%u %s does not exist", CMDSTATUS_BADQUESTION, arg);
 	return out;
     }
-    value = question_get_field(q, "", "description");
+    value = question_get_raw_field(q, "", "description");
     question_deref(q);
     if (value == NULL)
     {
@@ -831,6 +819,11 @@ command_data(struct confmodule *mod, char *arg)
             q->template = t;
             template_ref(t);
         }
+        else if (q->template != t) {
+            template_deref(q->template);
+            q->template = t;
+            template_ref(t);
+        }
         template_lset(t, NULL, item, value);
         mod->questions->methods.set(mod->questions, q);
         question_deref(q);
