@@ -208,12 +208,13 @@ static int is_confmodule(const char *filename)
  * Description: runs the given confmodule
  * Assumptions: none
  ************************************************************************/
-static int runconfmodule(int argc, char **argv)
+static int runconfmodule(const char *pkg, int argc, char **argv)
 {
 	struct confmodule *confmodule = NULL;
 	int ret;
 
 	confmodule = confmodule_new(g_config, g_templates, g_questions, g_frontend);
+	confmodule->owner = pkg;
 	confmodule->run(confmodule, argc, argv);
 	confmodule->communicate(confmodule);
 	confmodule->shutdown(confmodule);
@@ -293,7 +294,7 @@ static int runscript (const char *pkg, const char *script, const char *param) {
 		 */
 		argv[2] = (char *) param;
 		argv[3] = version;
-		if ((ret = runconfmodule(4, argv)) != 0)
+		if ((ret = runconfmodule(pkg, 4, argv)) != 0)
 		{
 			free(filename);
 			return ret;
@@ -301,6 +302,8 @@ static int runscript (const char *pkg, const char *script, const char *param) {
 	}
 	else
 	{
+		int filenamesize = strlen(filename) + strlen(param) + strlen(version) + 3;
+		filename = realloc(filename, filenamesize);
 		/* according to debconf:
 		 * Since a non confmodule might run other programs that
 		 * use debconf, checkpoint the db state and
@@ -311,7 +314,7 @@ static int runscript (const char *pkg, const char *script, const char *param) {
 	
 		unsetenv("DEBIAN_HAS_FRONTEND");
 
-		strvacat(filename, sizeof(filename), " ", param, " ", version, NULL);
+		strvacat(filename, filenamesize, " ", param, " ", version, NULL);
 		ret = system(filename);
 	
 		setenv("DEBIAN_HAS_FRONTEND", "1", 1);
