@@ -229,18 +229,21 @@ static int cdebconf_gtk_shutdown(struct frontend * fe)
 {
     struct frontend_data * fe_data = fe->data;
 
-    if (NULL != fe_data->event_listener) {
-        gtk_main_quit();
-        if (g_thread_self() != fe_data->event_listener) {
-            (void) g_thread_join(fe_data->event_listener);
+    if (NULL != fe_data)
+    {
+        if (NULL != fe_data->event_listener) {
+            gtk_main_quit();
+            if (g_thread_self() != fe_data->event_listener) {
+                (void) g_thread_join(fe_data->event_listener);
+            }
+            fe_data->event_listener = NULL;
         }
-        fe_data->event_listener = NULL;
-    }
-    cdebconf_gtk_destroy_main_window(fe);
+        cdebconf_gtk_destroy_main_window(fe);
 #ifdef DI_UDEB
-    cdebconf_gtk_di_shutdown(fe);
+        cdebconf_gtk_di_shutdown(fe);
 #endif /* DI_UDEB */
-    destroy_frontend_data(fe);
+        destroy_frontend_data(fe);
+    }
 
     return DC_OK;
 }
