@@ -3,8 +3,6 @@
  * cdebconf - An implementation of the Debian Configuration Management
  *            System
  *
- * $Id$
- *
  * cdebconf is (c) 2000-2007 Randolph Chung and others under the following
  * license.
  *
@@ -139,6 +137,48 @@ static void free_setters(struct frontend_data * fe_data)
     fe_data->setters = NULL;
 }
 
+/** Key event handler implementing the "Help" key shortcut.
+ *
+ * @param widget main widget
+ * @param key pressed key
+ * @param fe cdebconf frontend
+ * @return TRUE if "Help" shortcut has been pressed, FALSE otherwise
+ */
+static gboolean handle_help_key(GtkWidget * widget, GdkEventKey * key,
+                                struct frontend * fe)
+{
+    if (GDK_F1 == key->keyval || GDK_KP_F1 == key->keyval) {
+        cdebconf_gtk_help(fe);
+        return TRUE;
+    }
+    return FALSE;
+}
+
+/** Create the "Help" button.
+ *
+ * The button will be added to the main action box.
+ *
+ * @param fe cdebconf frontend
+ */
+static void create_help_button(struct frontend * fe)
+{
+    GtkWidget * button;
+    char * label;
+
+    label = cdebconf_gtk_get_text(fe, "debconf/button-help", "Help");
+    /* XXX: check NULL! */
+    button = gtk_button_new_with_label(label);
+    g_free(label);
+
+    g_signal_connect_swapped(G_OBJECT(button), "clicked",
+                             G_CALLBACK(cdebconf_gtk_help), fe);
+
+    cdebconf_gtk_add_button(fe, button);
+    cdebconf_gtk_set_button_secondary(fe, button, TRUE);
+    cdebconf_gtk_add_global_key_handler(
+        fe, button, G_CALLBACK(handle_help_key));
+}
+
 /** Key event handler implementing the "Go Back" key shortcut.
  *
  * @param widget main window
@@ -203,7 +243,7 @@ GtkWidget * cdebconf_gtk_create_continue_button(struct frontend * fe)
     gtk_container_set_focus_chain(GTK_CONTAINER(action_box), focus_chain);
     g_list_free(focus_chain);
 
-    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
+    gtk_widget_set_can_default(GTK_WIDGET(button), TRUE);
     gtk_widget_grab_default(GTK_WIDGET(button));
 
     return button;
@@ -346,7 +386,7 @@ static void update_question_database(struct frontend * fe)
 
     question = fe->questions;
     while (NULL != question) {
-        fe->qdb->methods.set(fe->qdb, question);
+        frontend_qdb_set(fe->qdb, question, 0);
         question = question->next;
     }
 }
@@ -358,13 +398,15 @@ static void update_question_database(struct frontend * fe)
  *
  * @param fe cdebconf frontend
  * @param question_box container for question handler widgets
- * @return FALSE if an error happened, TRUE otherwise
+ * @return DC_NOTIMPL if no handlers are available, DC_NOTOK if an error
+ *         happened, DC_OK otherwise
  */
-static gboolean call_question_handlers(struct frontend * fe,
-                                       GtkWidget * question_box)
+static int call_question_handlers(struct frontend * fe,
+                                 GtkWidget * question_box)
 {
     struct question * question;
     cdebconf_gtk_handler handler;
+    int ret;
 
     question = fe->questions;
     while (NULL != question) {
@@ -373,15 +415,15 @@ static gboolean call_question_handlers(struct frontend * fe,
             handler = find_external_handler(fe, question->template->type);
         }
         if (NULL == handler) {
-            return FALSE;
+            return DC_NOTIMPL;
         }
-        if (DC_OK != handler(fe, question, question_box)) {
+        if (DC_OK != (ret = handler(fe, question, question_box))) {
             g_warning("tag \"%s\" failed to display!", question->tag);
-            /* XXX: return FALSE? */
+            return ret;
         }
         question = question->next;
     }
-    return TRUE;
+    return DC_OK;
 }
 
 /** Create the container for question widgets in the given container.
@@ -416,7 +458,7 @@ static GtkWidget * create_question_box(struct frontend * fe,
     question_box = gtk_vbox_new(FALSE /* don't make children equal */,
                                 0 /* padding */);
 
-    if (IS_QUESTION_SINGLE(fe->questions)) {
+    if (0 && IS_QUESTION_SINGLE(fe->questions)) {
         gtk_box_pack_start(GTK_BOX(container), question_box,
                            TRUE /* expand */, TRUE /* fill */,
                            0 /* padding */);
@@ -477,7 +519,7 @@ static void wait_answer(struct frontend * fe)
  *    for the next frontend call.
  *
  * If a progress was running, the widgets are hidden at the beginning
- * of the procedure.
+ * of the procedure and shown at the end.
  *
  * @param fe cdebconf frontend
  * @return the debconf status corresponding to user actions
@@ -486,12 +528,15 @@ int cdebconf_gtk_go(struct frontend * fe)
 {
     struct frontend_data * fe_data = fe->data;
     GtkWidget * question_box;
+    struct question * question;
+    int ret;
 
     if (NULL == fe->questions) {
         return DC_OK;
     }
 
     cdebconf_gtk_set_answer(fe, DC_NO_ANSWER);
+    fe_data->help_question = NULL;
 
     gdk_threads_enter();
 #ifdef DI_UDEB
@@ -505,17 +550,36 @@ int cdebconf_gtk_go(struct frontend * fe)
     question_box = create_question_box(fe, fe_data->target_box);
     /* Hide the target box while we create the questions to avoid flicker. */
     cdebconf_gtk_hide_target_box(fe);
-    if (!call_question_handlers(fe, question_box)) {
-        /* an handler was missing */
-        cdebconf_gtk_set_answer(fe, DC_NOTIMPL);
-        goto end;
-    }
     if (CAN_GO_BACK(fe)) {
         create_goback_button(fe);
     }
+    if (DC_OK != (ret = call_question_handlers(fe, question_box))) {
+        cdebconf_gtk_set_answer(fe, ret);
+        gdk_threads_leave();
+        goto end;
+    }
     if (!is_action_box_filled(fe)) {
         create_default_buttons(fe);
     }
+
+    /* Set up the help button if necessary. FIXME: We only offer help for
+     * the first question with help text in any given question box. It's not
+     * clear what the right thing to do here is.
+     */
+    question = fe->questions;
+    while (NULL != question) {
+        char *help = q_get_help(fe, question);
+        if (help) {
+            struct question *help_q = fe->qdb->methods.get(fe->qdb, help);
+            if (help_q) {
+                fe_data->help_question = help_q;
+                create_help_button(fe);
+                break;
+            }
+        }
+        question = question->next;
+    }
+
     cdebconf_gtk_show_target_box(fe);
     cdebconf_gtk_show_buttons(fe);
     gdk_threads_leave();
@@ -535,9 +599,14 @@ int cdebconf_gtk_go(struct frontend * fe)
     }
     cdebconf_gtk_empty_target_box(fe);
     destroy_buttons(fe);
+    if (NULL != fe_data->progress_data) {
+        cdebconf_gtk_show_progress(fe);
+    }
     gdk_threads_leave();
 
 end:
+    question_deref(fe_data->help_question);
+    fe_data->help_question = NULL;
     free_setters(fe_data);
     return fe_data->answer;
 }
