@@ -3,8 +3,6 @@
  * cdebconf - An implementation of the Debian Configuration Management
  *            System
  *
- * $Id$
- *
  * cdebconf is (c) 2000-2007 Randolph Chung and others under the following
  * license.
  *
@@ -52,6 +50,7 @@
 #include "descriptions.h"
 #include "choice_model.h"
 #include "ui.h"
+#include "align_text_renderer.h"
 
 /** Setter function for the select handler in multiple questions form.
  *
@@ -102,19 +101,6 @@ static void set_value_from_select(struct question * question,
     }
 }
 
-/** Determine if a choice is top-level for countrychooser/country-name.
- *
- * @param index choice index
- * @param choice choice canonical value
- * @param choice_translated choice translated value
- * @return TRUE if top-level, FALSE otherwise
- */
-static gboolean is_country(int index, const char * choice,
-                           const char * choice_translated)
-{
-    return '-' == choice_translated[0] && '-' == choice_translated[1];
-}
-
 /** Determine if a choice is top-level for partman/choose_partition.
  *
  * @param index choice index
@@ -125,7 +111,7 @@ static gboolean is_country(int index, const char * choice,
 static gboolean is_disk(int index, const char * choice,
                         const char * choice_translated)
 {
-    return NULL == strstr(choice_translated, "    ");
+    return '>' != choice_translated[0];
 }
 
 /** List of predicates used for questions renderered as tree. */
@@ -137,7 +123,6 @@ static const struct {
     /** expand all parents */
     gboolean expand_all;
 } special_questions[] = {
-    { "countrychooser/country-name", is_country, FALSE },
     { "partman/choose_partition", is_disk, TRUE },
     { NULL, NULL, FALSE }
 };
@@ -227,6 +212,122 @@ static gboolean focus_path(GtkTreeView * view, GdkEventExpose * event,
     return FALSE; /* propagate the event */
 }
 
+/** Private structure for adjust_tabs and adjust_tabs_for_choice. */
+struct adjust_data {
+    /** Widget where the tabs will be renderered. */
+    GtkWidget * widget;
+    /** Tab array to update that will be updated. */
+    PangoTabArray * tab_array;
+};
+
+/** Adjust tab stops for the given node to display all the given values.
+ *
+ * @param model model being handled
+ * @param path path of the current node
+ * @param iter iter of the current node
+ * @param adjust_data data prepared by adjust_tabs
+ * @return always FALSE (= never stop)
+ */
+static gboolean adjust_tabs_for_node(GtkTreeModel * model, GtkTreePath * path,
+                                     GtkTreeIter * iter,
+                                     struct adjust_data * adjust_data)
+{
+    char * choice;
+    char ** values;
+    gint columns;
+    gint value_width;
+    gint previous_location;
+    gint location;
+    gint i;
+
+    gtk_tree_model_get(model, iter,
+        /* column: */ CHOICE_MODEL_TRANSLATED_VALUE, &choice,
+        -1 /* end of list */);
+    values = g_strsplit(choice, "\t", 0 /* split all */);
+
+    columns = g_strv_length(values);
+    if (pango_tab_array_get_size(adjust_data->tab_array) < columns - 1) {
+        pango_tab_array_resize(adjust_data->tab_array, columns - 1);
+    }
+    previous_location = 0;
+    for (i = 0; columns > i + 1; i++) {
+        value_width = cdebconf_gtk_get_text_width(
+                          adjust_data->widget, values[i]) + COLUMN_SPACING;
+        pango_tab_array_get_tab(adjust_data->tab_array, i,
+                                NULL /* don't get alignment */, &location);
+        if (location - previous_location < value_width) {
+            location = previous_location + value_width;
+            pango_tab_array_set_tab(adjust_data->tab_array, i, PANGO_TAB_LEFT,
+                                    location);
+        }
+        previous_location = location;
+    }
+
+#if 0
+    /* DEBUG: dump tabs */
+    g_warning("dump after choice: %s", choice);
+    for (i = 0; pango_tab_array_get_size(adjust_data->tab_array) > i; i++) {
+        pango_tab_array_get_tab(adjust_data->tab_array, i,
+                                NULL /* don't get alignment */,
+                                &location);
+        g_warning("%d: %d", i, location);
+    }
+#endif
+
+    g_free(choice);
+    g_strfreev(values);
+
+    return FALSE;
+}
+
+/** Adjust tab stops in the given tab array to render all the "columns"
+ * (separated by a tab) in the translated choice of the given model.
+ *
+ * @param widget widget where the choices will be rendered
+ * @param tab_array tab array to update
+ * @param model a choice model
+ */
+static void adjust_tabs(GtkWidget * widget, PangoTabArray * tab_array,
+                        GtkTreeModel * model)
+{
+    struct adjust_data adjust_data;
+
+    adjust_data.widget = widget;
+    adjust_data.tab_array = tab_array;
+    gtk_tree_model_foreach(
+        model, (GtkTreeModelForeachFunc) adjust_tabs_for_node, &adjust_data);
+}
+
+/** Insert the column displaying translated choices in the given GtkTreeView.
+ *
+ * @param fe frontend
+ * @param view column destination
+ */
+static void insert_choice_column(struct frontend * fe, GtkTreeView * view)
+{
+    GtkCellRenderer * renderer;
+    PangoTabArray * tab_array;
+
+    if (CAN_ALIGN(fe)) {
+        /* XXX: check NULL */
+        tab_array = pango_tab_array_new(0 /* start with no tabs */,
+                                        FALSE /* use pango unit */);
+        adjust_tabs(GTK_WIDGET(view), tab_array,
+                    gtk_tree_view_get_model(view));
+        renderer = cdebconf_gtk_align_text_renderer_new();
+        cdebconf_gtk_align_text_renderer_set_tab_array(
+            ALIGN_TEXT_RENDERER(renderer), tab_array);
+        pango_tab_array_free(tab_array);
+    } else {
+        renderer = gtk_cell_renderer_text_new();
+    }
+    gtk_tree_view_insert_column_with_attributes(
+        view, -1 /* insert at the end */,
+        NULL /* no title */, renderer,
+        "text", CHOICE_MODEL_TRANSLATED_VALUE,
+        NULL /* end of attribute list */);
+}
+
 /** Create widget for select question in single question form.
  *
  * This will also register the corresponding setter function.
@@ -245,8 +346,6 @@ static int create_select_list(struct frontend * fe, struct question * question,
     GtkWidget * view;
     GtkWidget * scroll;
     GtkWidget * frame;
-    GtkCellRenderer * text_renderer;
-    char * description;
 
     /* check NULL! */
     view = gtk_tree_view_new_with_model(model);
@@ -265,14 +364,7 @@ static int create_select_list(struct frontend * fe, struct question * question,
         hide_expanders(GTK_TREE_VIEW(view));
     }
 
-    description = q_get_description(question);
-    text_renderer = gtk_cell_renderer_text_new();
-    gtk_tree_view_insert_column_with_attributes(
-        GTK_TREE_VIEW(view), -1 /* insert at the end */,
-        description /* title */, text_renderer,
-        "text", CHOICE_MODEL_TRANSLATED_VALUE,
-        NULL /* end of attribute list */);
-    g_free(description);
+    insert_choice_column(fe, GTK_TREE_VIEW(view));
 
     g_signal_connect_swapped(G_OBJECT(view), "row-activated",
                              G_CALLBACK(cdebconf_gtk_set_answer_ok), fe);
@@ -440,6 +532,26 @@ static gulong connect_signal_to_row(GObject * object, const gchar * signal,
                                  0 /* no connect options */);
 }
 
+/** Handler for the "cursor-changed" signal that forces the cursor to stay on
+ * the first column.
+ *
+ * This will disallow the selection of option descriptions during multiselect.
+ *
+ * @param view the tree view
+ * @param fe cdebconf frontend
+ */
+static void handle_cursor_changed(GtkTreeView * view,
+                                  struct frontend * fe)
+{
+    GtkTreePath * path;
+
+    gtk_tree_view_get_cursor(view, &path, NULL);
+    g_signal_handlers_block_by_func(view, handle_cursor_changed, fe);
+    gtk_tree_view_set_cursor(view, path, gtk_tree_view_get_column(view, 0),
+                             FALSE /* no editing */);
+    g_signal_handlers_unblock_by_func(view, handle_cursor_changed, fe);
+}
+
 /** Create widget for multiselect question in single question form.
  *
  * This will also register the corresponding setter function.
@@ -460,7 +572,6 @@ static int create_multiselect_list(struct frontend * fe,
     GtkWidget * scroll;
     GtkWidget * frame;
     GtkCellRenderer * toggle_renderer;
-    GtkCellRenderer * text_renderer;
     GtkTreePath * path;
 
     view = gtk_tree_view_new_with_model(model);
@@ -476,17 +587,15 @@ static int create_multiselect_list(struct frontend * fe,
         "active", CHOICE_MODEL_SELECTED,
         NULL /* end of attribute list */);
 
-    text_renderer = gtk_cell_renderer_text_new();
-    gtk_tree_view_insert_column_with_attributes(
-        GTK_TREE_VIEW(view), -1 /* insert at the end */,
-        NULL /* no title */, text_renderer,
-        /* column: */ "text", CHOICE_MODEL_TRANSLATED_VALUE,
-        NULL /* end of list */);
+    insert_choice_column(fe, GTK_TREE_VIEW(view));
 
     if (!IS_SPECIAL_QUESTION(question)) {
         hide_expanders(GTK_TREE_VIEW(view));
     }
 
+    g_signal_connect(G_OBJECT(view), "cursor-changed",
+                     G_CALLBACK(handle_cursor_changed), fe);
+
     /* select the first row */
     gtk_tree_model_get_iter_first(model, &iter);
     path = gtk_tree_model_get_path(model, &iter);
