@@ -1115,7 +1115,7 @@ newt_help_callback(newtComponent co, void *d)
 }
 
 /* ----------------------------------------------------------------------- */
-struct question_handlers {
+static struct question_handlers {
 	const char *type;
 	newt_handler *handler;
 } question_handlers[] = {
@@ -1220,7 +1220,7 @@ newt_go(struct frontend *obj)
                 ret = handler(obj, q);
                 newtPopHelpLine();
                 if (ret == DC_OK)
-                    obj->qdb->methods.set(obj->qdb, q);
+                    frontend_qdb_set(obj->qdb, q, 0);
                 if (*help) {
                     newtSetHelpCallback(NULL);
                     help_cb_data = NULL;
