@@ -10,8 +10,6 @@
  * friendly implementation. I've taken care to make the prompts work well
  * with screen readers and the like.
  *
- * $Id: text.c 15039 2004-05-09 07:01:22Z barbier $
- *
  * cdebconf is (c) 2000-2001 Randolph Chung and others under the following
  * license.
  *
@@ -44,6 +42,7 @@
 #include "question.h"
 #include "frontend.h"
 #include "database.h"
+#include "plugin.h"
 #include "strutl.h"
 
 #include <ctype.h>
@@ -59,11 +58,11 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
-#define q_get_extended_description(q)   question_get_field((q), "", "extended_description")
-#define q_get_description(q)		question_get_field((q), "", "description")
-#define q_get_choices(q)		question_get_field((q), "", "choices")
-#define q_get_choices_vals(q)		question_get_field((q), NULL, "choices")
-#define q_get_indices(q)		question_get_field((q), "", "indices")
+struct frontend_data {
+	char *previous_title;
+};
+
+typedef int (text_handler)(struct frontend *obj, struct question *q);
 
 #define MAKE_UPPER(C) do { if (islower((int) C)) { C = (char) toupper((int) C); } } while(0)
 
@@ -77,30 +76,6 @@
 #define ISEMPTY(buf) (buf[0] == 0)
 #endif
 
-/*  This function will be moved to strutl.c  */
-/*
- * Add spaces at the end of string so that strwidth(that) == maxsize
- * Input string must have been allocated with enough space.
- */
-int
-strpad (char *what, size_t maxsize)
-{
-    size_t pos;
-    int k;
-    char *p;
-    wchar_t c;
-
-    pos = 0;
-    for (p = what; (k = mbtowc (&c, p, MB_LEN_MAX)) > 0; p += k)
-        pos += wcwidth (c);
-    if (pos > maxsize)
-        return 0;
-    for (k = pos; k < maxsize; k++, p++)
-        *p = ' ';
-    *p = '\0';
-    return 1;
-}  
-
 /*
  * Function: getwidth
  * Input: none
@@ -108,7 +83,7 @@ strpad (char *what, size_t maxsize)
  * Description: get the width of the current terminal
  * Assumptions: doesn't handle resizing; caches value on first call
  */
-static const int getwidth(void)
+static int getwidth(void)
 {
 	static int res = 80;
 	static int inited = 0;
@@ -128,22 +103,6 @@ static const int getwidth(void)
 	return res;
 }
 
-/*
- * Function: get_text
- * Input: struct frontend *obj - frontend object
- *        const char *template - template name
- *        const char *fallback - string to use if not available
- * Output: const char * - ptr to string, translated if possible
- * Description: get the translated version of a string
- * Assumptions: None.
- */
-static const char *
-get_text(struct frontend *obj, const char *template, const char *fallback)
-{
-	struct question *q = obj->qdb->methods.get(obj->qdb, template);
-	return q ? q_get_description(q) : fallback;
-}
-
 /*
  * Function: wrap_print
  * Input: const char *str - string to display
@@ -176,13 +135,13 @@ static void wrap_print(const char *str)
  */
 static void text_handler_displaydesc(struct frontend *obj, struct question *q) 
 {
-	char *descr = q_get_description(q);
-	char *ext_descr = q_get_extended_description(q);
+	char *descr = q_get_description(obj, q);
+	char *ext_descr = q_get_extended_description(obj, q);
 	if (strcmp(q->template->type, "note") == 0 ||
 	    strcmp(q->template->type, "error") == 0)
 	{
 		if (strcmp(q->template->type, "error") == 0)
-			printf(get_text(obj, "debconf/text-error", "!! ERROR: %s"), descr);
+			printf(question_get_text(obj, "debconf/text-error", "!! ERROR: %s"), descr);
 		else
 			printf("%s", descr);
 		printf("\n\n");
@@ -209,19 +168,37 @@ get_answer(char *answer, int size)
 static void
 show_help (struct frontend *obj, struct question *q)
 {
-	char *descr = q_get_description(q);
-	printf("%s", get_text(obj, "debconf/text-help-keystrokes", "KEYSTROKES:"));
+	char *descr = q_get_description(obj, q);
+	char *help = q_get_help(obj, q);
+	if (*help) {
+		struct question *help_q = obj->qdb->methods.get(obj->qdb, help);
+		if (help_q) {
+			char *help_descr = q_get_description(obj, help_q);
+			char *help_ext_descr = q_get_extended_description(obj, help_q);
+			wrap_print(help_descr);
+			printf("\n");
+			if (*help_ext_descr) {
+				wrap_print(help_ext_descr);
+				printf("\n");
+			}
+			free(help_ext_descr);
+			free(help_descr);
+			question_deref(help_q);
+		}
+		free(help);
+	}
+	printf("%s", question_get_text(obj, "debconf/text-help-keystrokes", "KEYSTROKES:"));
 	printf("\n  %c  ", CHAR_HELP);
-	printf("%s", get_text(obj, "debconf/text-help-help", "Display this help message"));
+	printf("%s", question_get_text(obj, "debconf/text-help-help", "Display this help message"));
 	if (obj->methods.can_go_back (obj, q)) {
 		printf("\n  %c  ", CHAR_GOBACK);
-		printf("%s", get_text(obj, "debconf/text-help-goback", "Go back to previous question"));
+		printf("%s", question_get_text(obj, "debconf/text-help-goback", "Go back to previous question"));
 	}
 	if (strcmp(q->template->type, "string") == 0 ||
 	    strcmp(q->template->type, "passwd") == 0 ||
 	    strcmp(q->template->type, "multiselect") == 0) {
 		printf("\n  %c  ", CHAR_CLEAR);
-		printf("%s", get_text(obj, "debconf/text-help-clear", "Select an empty entry"));
+		printf("%s", question_get_text(obj, "debconf/text-help-clear", "Select an empty entry"));
 	}
 	printf("\n");
 	wrap_print(descr);
@@ -247,12 +224,18 @@ printlist (struct frontend *obj, struct question *q, int count, char **choices_t
 		logcount++;
 	} while (i > 0);
 
+	if (obj->methods.can_align(obj, q)) {
+		stralign(choices_translated, count);
+		/* Display only one column to get meaningful alignment. */
+		width = 1;
+	}
+
 	/*  Set string arrays  */
 	for (i=0; i < count; i++)
 	{
 		/*  Trailing spaces are a placeholder to add [*] for
 		    selected values */
-		asprintf(&(fchoices[i]), "  %d. %s    ", i+1, choices_translated[i]);
+		asprintf(&(fchoices[i]), "%3d. %s    ", i+1, choices_translated[i]);
 		if (selected[tindex[i]])
 			strcpy(fchoices[i]+strlen(fchoices[i])-3, "[*]");
 		if (strwidth(fchoices[i]) < choice_min || choice_min == -1)
@@ -356,14 +339,14 @@ static int text_handler_boolean(struct frontend *obj, struct question *q)
 	}
 
 	do {
-		printf("  %d. %s%s", 1, get_text(obj, "debconf/yes", "Yes"), (1 == def ? " [*]" : "    "));
-		printf("  %d. %s%s", 2, get_text(obj, "debconf/no", "No"), (2 == def ? " [*]" : ""));
+		printf("  %d. %s%s", 1, question_get_text(obj, "debconf/yes", "Yes"), (1 == def ? " [*]" : "    "));
+		printf("  %d. %s%s", 2, question_get_text(obj, "debconf/no", "No"), (2 == def ? " [*]" : ""));
 		printf("\n");
 		if (def)
-			printf(get_text(obj, "debconf/text-prompt-default",
+			printf(question_get_text(obj, "debconf/text-prompt-default",
 					"Prompt: '%c' for help, default=%d> "), CHAR_HELP, def);
 		else
-			printf(get_text(obj, "debconf/text-prompt",
+			printf(question_get_text(obj, "debconf/text-prompt",
 					"Prompt: '%c' for help> "), CHAR_HELP);
 		get_answer(buf, sizeof(buf));
 		if (buf[0] == CHAR_HELP && buf[1] == 0)
@@ -403,15 +386,19 @@ static int text_handler_multiselect(struct frontend *obj, struct question *q)
 	int i, j, count = 0, dcount, choice;
 	int *tindex = NULL;
 	int ret = DC_OK;
-	const char *indices = q_get_indices(q);
+	const char *indices = q_get_indices(obj, q);
 
-	count = strgetargc(q_get_choices_vals(q));
+	count = strgetargc(q_get_choices_vals(obj, q));
 	if (count <= 0)
 		return DC_NOTOK;
 	choices = malloc(sizeof(char *) * count);
 	choices_translated = malloc(sizeof(char *) * count);
+	for (i = 0; i < count; ++i)
+	{
+		choices[i] = choices_translated[i] = NULL;
+	}
 	tindex = malloc(sizeof(int) * count);
-	if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_translated, tindex, count) != count)
+	if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_translated, tindex, count) != count)
 	{
 		ret = DC_NOTOK;
 		goto CleanUp_TINDEX;
@@ -443,7 +430,7 @@ static int text_handler_multiselect(struct frontend *obj, struct question *q)
 
   DISPLAY:
 	printlist (obj, q, count, choices_translated, tindex, selected);
-	printf(get_text(obj, "debconf/text-prompt-default-string", 
+	printf(question_get_text(obj, "debconf/text-prompt-default-string", 
 		"Prompt: '%c' for help, default=%s> "), CHAR_HELP, defval);
 	get_answer(answer, sizeof(answer));
 	if (answer[0] == CHAR_HELP && answer[1] == 0)
@@ -526,15 +513,19 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 	const char *defval = question_getvalue(q, "");
 	int *tindex = NULL;
 	int ret = DC_OK;
-	const char *indices = q_get_indices(q);
+	const char *indices = q_get_indices(obj, q);
 
-	count = strgetargc(q_get_choices_vals(q));
+	count = strgetargc(q_get_choices_vals(obj, q));
 	if (count <= 0)
 		return DC_NOTOK;
 	choices = malloc(sizeof(char *) * count);
 	choices_translated = malloc(sizeof(char *) * count);
+	for (i = 0; i < count; ++i)
+	{
+		choices[i] = choices_translated[i] = NULL;
+	}
 	tindex = malloc(sizeof(int) * count);
-	if (strchoicesplitsort(q_get_choices_vals(q), q_get_choices(q), indices, choices, choices_translated, tindex, count) != count)
+	if (strchoicesplitsort(q_get_choices_vals(obj, q), q_get_choices(obj, q), indices, choices, choices_translated, tindex, count) != count)
 	{
 		ret = DC_NOTOK;
 		goto CleanUp_TINDEX;
@@ -559,11 +550,11 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 	do {
 		printlist (obj, q, count, choices_translated, tindex, selected);
 		if (def >= 0 && choices_translated[def]) {
-			printf(get_text(obj, "debconf/text-prompt-default", 
+			printf(question_get_text(obj, "debconf/text-prompt-default", 
 				"Prompt: '%c' for help, default=%d> "),
 					CHAR_HELP, def+1);
 		} else {
-			printf(get_text(obj, "debconf/text-prompt",
+			printf(question_get_text(obj, "debconf/text-prompt",
 				"Prompt: '%c' for help> "), CHAR_HELP);
 		}
 		get_answer(answer, sizeof(answer));
@@ -612,8 +603,9 @@ static int text_handler_select(struct frontend *obj, struct question *q)
 static int text_handler_note(struct frontend *obj, struct question *q)
 {
 	char buf[100] = {0};
-	printf("%s\n", get_text(obj, "debconf/cont-prompt",
-				"[Press enter to continue]"));
+	printf("%s ", question_get_text(obj, "debconf/cont-prompt",
+			       "[Press enter to continue]"));
+	fflush(stdout);
 	while (1)
 	{
 		get_answer(buf, sizeof(buf));
@@ -695,9 +687,9 @@ static int text_handler_string(struct frontend *obj, struct question *q)
 	const char *defval = question_getvalue(q, "");
 	while (1) {
 		if (defval)
-			printf(get_text(obj, "debconf/text-prompt-default-string", "Prompt: '%c' for help, default=%s> "), CHAR_HELP, defval);
+			printf(question_get_text(obj, "debconf/text-prompt-default-string", "Prompt: '%c' for help, default=%s> "), CHAR_HELP, defval);
 		else
-			printf(get_text(obj, "debconf/text-prompt", "Prompt: '%c' for help> "), CHAR_HELP);
+			printf(question_get_text(obj, "debconf/text-prompt", "Prompt: '%c' for help> "), CHAR_HELP);
 		fflush(stdout);
 		get_answer(buf, sizeof(buf));
 		if (buf[0] == CHAR_HELP && buf[1] == 0)
@@ -746,9 +738,9 @@ static int text_handler_error(struct frontend *obj, struct question *q)
 }
 
 /* ----------------------------------------------------------------------- */
-struct question_handlers {
+static struct question_handlers {
 	const char *type;
-	int (*handler)(struct frontend *obj, struct question *q);
+	text_handler *handler;
 } question_handlers[] = {
 	{ "boolean",	text_handler_boolean },
 	{ "multiselect", text_handler_multiselect },
@@ -758,10 +750,11 @@ struct question_handlers {
 	{ "string",	text_handler_string },
 	{ "text",	text_handler_text },
 	{ "error",	text_handler_error },
+	{ "",		NULL },
 };
 
 /*
- * Function: text_intitialize
+ * Function: text_initialize
  * Input: struct frontend *obj - frontend UI object
  *        struct configuration *cfg - configuration parameters
  * Output: int - DC_OK, DC_NOTOK
@@ -773,8 +766,32 @@ struct question_handlers {
  */
 static int text_initialize(struct frontend *obj, struct configuration *conf)
 {
+	struct frontend_data *data = NEW(struct frontend_data);
+	char *term = getenv("TERM");
+	char *palette = getenv("FRONTEND_BACKGROUND");
+	data->previous_title = NULL;
+	obj->data = data;
 	obj->interactive = 1;
 	signal(SIGINT, SIG_IGN);
+	if (palette && !strcmp(palette, "dark") &&
+			term && (!strcmp(term, "linux") || !strcmp(term, "bterm"))) {
+		/* Hard-code for these cases */
+		printf("\e[37m\e[40m\e[1m\e[H\e[J");
+		fflush(stdout);
+	}
+	return DC_OK;
+}
+
+static int text_shutdown(struct frontend *obj)
+{
+	struct frontend_data *data = (struct frontend_data *)obj->data;
+	if (NULL != data)
+	{
+		if (NULL != data->previous_title)
+			free(data->previous_title);
+		free(data);
+		obj->data = NULL;
+	}
 	return DC_OK;
 }
 
@@ -786,12 +803,50 @@ static int text_initialize(struct frontend *obj, struct configuration *conf)
  * Description: tells whether confmodule supports backing up
  * Assumptions: none
  */
-static int
+static bool
 text_can_go_back(struct frontend *obj, struct question *q)
 {
 	return (obj->capability & DCF_CAPB_BACKUP);
 }
 
+/*
+ * Function: text_can_align
+ * Input: struct frontend *obj - frontend object
+ *        struct question *q - question object
+ * Output: int - DC_OK, DC_NOTOK
+ * Description: tells whether confmodule supports aligning columns
+ * Assumptions: none
+ */
+static bool
+text_can_align(struct frontend *obj, struct question *q)
+{
+	return (obj->capability & DCF_CAPB_ALIGN);
+}
+
+/*
+ * Function: text_lookup_directive
+ * Input: struct frontend *obj - frontend object
+ *        const char *directive - the directive
+ * Output: const char * - directive value or NULL
+ * Description: returns directive value in the given context
+ * Assumptions: none
+ */
+static const char *
+text_lookup_directive(struct frontend *obj, const char *directive)
+{
+	if (obj->methods.can_align(obj, obj->questions)) {
+		if (strcmp("TAB", directive) == 0) {
+			return STRALIGN_TAB;
+		} else if (strcmp("ALIGN=CENTER", directive) == 0) {
+			return STRALIGN_ALIGN_CENTER;
+		} else if (strcmp("ALIGN=RIGHT", directive) == 0) {
+			return STRALIGN_ALIGN_RIGHT;
+		}
+	}
+	/* Remove unhandled directives */
+	return "";
+}
+
 /*
  * Function: text_go
  * Input: struct frontend *obj - frontend object
@@ -801,51 +856,97 @@ text_can_go_back(struct frontend *obj, struct question *q)
  */
 static int text_go(struct frontend *obj)
 {
+	struct frontend_data *data = (struct frontend_data *) obj->data;
 	struct question *q = obj->questions;
 	int i;
 	int ret = DC_OK;
-	int display_title = 1;
 
 	while (q != NULL) {
 		for (i = 0; i < DIM(question_handlers); i++) {
-			if (strcmp(q->template->type, question_handlers[i].type) == 0) 
-			{
+			text_handler *handler;
+			struct plugin *plugin = NULL;
 
-				if (display_title)
+			if (*question_handlers[i].type)
+				handler = question_handlers[i].handler;
+			else {
+				plugin = plugin_find(obj, q->template->type);
+				if (plugin) {
+					INFO(INFO_DEBUG, "Found plugin for %s", q->template->type);
+					handler = (text_handler *) plugin->handler;
+				} else {
+					INFO(INFO_DEBUG, "No plugin for %s", q->template->type);
+					continue;
+				}
+			}
+
+			if (plugin || strcmp(q->template->type, question_handlers[i].type) == 0)
+			{
+				if (!data->previous_title ||
+				    strcmp(obj->title, data->previous_title) != 0)
 				{
-					printf("%s\n\n", obj->title);
-					display_title = 0;
+					size_t underline_len;
+					char *underline;
+
+					/* TODO: can't tell if we called go()
+					 * twice during one progress bar, but
+					 * I'm guessing that's relatively
+					 * unimportant (c.f. #271707).
+					 */
+					if (obj->progress_title != NULL)
+						putchar('\n');
+					underline_len = strlen(obj->title);
+					underline = malloc(underline_len + 1);
+					memset(underline, '-', underline_len);
+					underline[underline_len] = '\0';
+					printf("%s\n%s\n\n", obj->title, underline);
+					free(underline);
+					if (data->previous_title)
+						free(data->previous_title);
+					data->previous_title = strdup(obj->title);
 				}
 				text_handler_displaydesc(obj, q);
-				ret = question_handlers[i].handler(obj, q);
+				ret = handler(obj, q);
+				putchar('\n');
 				if (ret == DC_OK)
-					obj->qdb->methods.set(obj->qdb, q);
-				else if (ret == DC_GOBACK && q->prev != NULL)
-					q = q->prev;
-				else
-					return ret;
+					frontend_qdb_set(obj->qdb, q, 0);
+				if (plugin)
+					plugin_delete(plugin);
 				break;
 			}
 		}
+		if (ret == DC_NOTOK)
+			break;
+		if (i == DIM(question_handlers))
+			return DC_NOTIMPL;
 		if (ret == DC_OK)
 			q = q->next;
+		else if (ret == DC_GOBACK) {
+			do {
+			    q = q->prev;
+			} while (q != NULL && 0 == strcmp("error", q->template->type));
+		}
 	}
-	return DC_OK;
+	return ret;
 }
 
-static void text_progress_start(struct frontend *obj, int min, int max, const char *title)
+static void text_progress_start(struct frontend *obj, int min, int max, struct question *title)
 {
-	DELETE(obj->progress_title);
-	obj->progress_title = STRDUP(title);
+	char *title_desc;
+
+	question_deref(obj->progress_title);
+	obj->progress_title = title;
+	question_ref(obj->progress_title);
 	obj->progress_min = min;
 	obj->progress_max = max;
 	obj->progress_cur = min;
 
-	printf("%s  ", title);
+	title_desc = q_get_raw_description(title);
+	printf("%s  ", title_desc);
+	free(title_desc);
 	fflush(stdout);
 }
 
-static void text_progress_set(struct frontend *obj, int val)
+static int text_progress_set(struct frontend *obj, int val)
 {
 	static int last = 0;
 	int new;
@@ -857,23 +958,30 @@ static void text_progress_set(struct frontend *obj, int val)
 		last = 0;
 	/*  Prevent verbose output  */
 	if (new / 10 == last / 10)
-		return;
+		return DC_OK;
 	last = new;
 	printf("..%d%%", new);
 	fflush(stdout);
+
+	return DC_OK;
 }
 
 static void text_progress_stop(struct frontend *obj)
 {
-	INFO(INFO_DEBUG, "%s\n", __FUNCTION__);
+	INFO(INFO_DEBUG, "%s", __FUNCTION__);
 	printf("\n");
 	fflush(stdout);
+	question_deref(obj->progress_title);
+	obj->progress_title = NULL;
 }
 
 struct frontend_module debconf_frontend_module =
 {
 	initialize: text_initialize,
+	shutdown: text_shutdown,
+	lookup_directive: text_lookup_directive,
 	can_go_back: text_can_go_back,
+	can_align: text_can_align,
 	go: text_go,
 	progress_start: text_progress_start,
 	progress_set: text_progress_set,
